/*
 * ToAutomatonDerivation.cpp
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#include "ToAutomatonDerivation.h"

#include <alib/set>
#include <alib/deque>
#include <alib/vector>
#include <alib/hexavigesimal>

#include <string/LinearString.h>

#include <regexp/transform/RegExpDerivation.h>
#include <regexp/simplify/RegExpOptimize.h>
#include <regexp/properties/RegExpEpsilon.h>
#include <registration/AlgoRegistration.hpp>

namespace regexp {

namespace convert {

template<class T>
automaton::DFA<> ToAutomatonDerivation::convert(const T& regexp) {
	// 1.
	T V = regexp::simplify::RegExpOptimize::optimize(regexp);

	ext::set<T> Q = { V };
	ext::deque<ext::set<T>> Qi;

	Qi.push_back(ext::set<T>());
	Qi.at(0).insert(V);

	int i = 1;

	// 2.
	while(! Qi.at(i - 1).empty()) {
		Qi.push_back(ext::set<T>()); // initialize set Q_i

		for(const auto& dregexp : Qi.at(i - 1)) {

			for(const auto& a : regexp.getAlphabet()) {
				string::LinearString < > string(ext::vector<DefaultSymbolType>{a});
				T derived = regexp::RegExpDerivation::derivation(dregexp, string);
				derived = regexp::simplify::RegExpOptimize::optimize(derived);

				// this will also add \emptyset as a regexp (and as FA state)
				if(Q.count(derived) == 0) // if this state has already been found, do not add
					Qi.at(i).insert(derived);
			}
		}

		Q.insert(Qi.at(i).begin(), Qi.at(i).end());
		i += 1;
	}

	ext::map<T, DefaultStateType> stateMap;
	int stateId = 0;

	for(const auto& r : Q) {
		DefaultStateType q = DefaultStateType (ext::toBase26(stateId++));
		stateMap.insert(std::make_pair(r, q));
	}

	// ------------------------------------------------------------------------
	// 3.

	automaton::DFA<> automaton(stateMap.find(V)->second);

	for(const auto& r : stateMap) {
		automaton.addState(r.second);
	}

	automaton.setInputAlphabet(regexp.getAlphabet());

	for(const auto& r : Q) {

		for(const auto& a: regexp.getAlphabet()) {
			string::LinearString < > string(ext::vector<DefaultSymbolType>{a});
			T derived = regexp::RegExpDerivation::derivation(r, string);
			derived = regexp::simplify::RegExpOptimize::optimize(derived);

			automaton.addTransition(stateMap.find(r)->second, a, stateMap.find(derived)->second);
		}
	}

	for(const auto& r : Q) {
		if(regexp::properties::RegExpEpsilon::languageContainsEpsilon(r))
			automaton.addFinalState(stateMap.find(r)->second);
	}

	return automaton;
}

auto ToAutomatonDerivationFormalRegExp = registration::AbstractRegister < ToAutomatonDerivation, automaton::DFA < >, const regexp::FormalRegExp < > & > ( ToAutomatonDerivation::convert );
auto ToAutomatonDerivationUnboundedRegExp = registration::AbstractRegister < ToAutomatonDerivation, automaton::DFA < >, const regexp::UnboundedRegExp < > & > ( ToAutomatonDerivation::convert );

} /* namespace convert */

} /* namespace regexp */
