/*
 * RightLG.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef RIGHT_LG_H_
#define RIGHT_LG_H_

#include <alib/map>
#include <alib/set>
#include <alib/vector>
#include <alib/variant>
#include <alib/algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../common/GrammarNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace grammar {

class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

/**
 * \brief
 * Right linear grammar in Chomsky hierarchy or type 3 in Chomsky hierarchy. Generates regular languages.

 * \details
 * Definition is similar to all common definitions of 'relaxed' regular grammars.
 * G = (N, T, P, S),
 * N (NonterminalAlphabet) = nonempty finite set of nonterminal symbols,
 * T (TerminalAlphabet) = finite set of terminal symbols - having this empty won't let grammar do much though,
 * P = set of production rules of the form A -> Ba or A -> a, where A, B \in N and a \in T*,
 * S (InitialSymbol) = initial nonterminal symbol,
 *
 * \tparam SymbolType used for the terminal alphabet, the nonterminal alphabet, and the initial symbol of the grammar.
 */
template < class SymbolType >
class RightLG final : public GrammarBase, public core::Components < RightLG < SymbolType >, ext::set < SymbolType >, component::Set, std::tuple < TerminalAlphabet, NonterminalAlphabet >, SymbolType, component::Value, InitialSymbol > {
	/**
	 * Rules as mapping from nonterminal symbol on the left hand side to set of either sequence of terminal symbols or doublets of sequence of nonterminal symbol and terminal symbols.
	 */
	ext::map < SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > rules;

public:
	/**
	 * \brief Creates a new instance of Right linear grammar with a concrete initial symbol.
	 *
	 * \param initialSymbol the initial symbol of the grammar
	 */
	explicit RightLG ( SymbolType initialSymbol );

	/**
	 * \brief Creates a new instance of Right linear grammar with a concrete nonterminal, terminal alphabet and initial symbol.
	 *
	 * \param nonTerminalSymbols the initial nonterminal alphabet
	 * \param terminalSymbols the initial terminal alphabet
	 * \param initialSymbol the initial symbol of the grammar
	 */
	explicit RightLG ( ext::set < SymbolType > nonTerminalSymbols, ext::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	/**
	 * @copydoc grammar::GrammarBase::clone()
	 */
	virtual GrammarBase * clone ( ) const override;

	/**
	 * @copydoc grammar::GrammarBase::plunder()
	 */
	virtual GrammarBase * plunder ( ) && override;

	/**
	 * \brief Add a new rule of a grammar.
	 *
	 * \details The rule is in a form of A -> Ba or A -> a, where A, B \in N and a \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed added, false othervise
	 */
	bool addRule ( SymbolType leftHandSide, ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > rightHandSide );

	/**
	 * \brief Add a new rule of a grammar.
	 *
	 * \details The rule is in a form of A -> a, where A \in N and a \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed added, false othervise
	 */
	bool addRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide );

	/**
	 * \brief Add a new rule of a grammar.
	 *
	 * \details The rule is in a form of A -> Ba, where A, B \in N and a \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed added, false othervise
	 */
	bool addRule ( SymbolType leftHandSide, ext::pair < ext::vector < SymbolType >, SymbolType > rightHandSide );

	/**
	 * \brief Add new rules of a grammar.
	 *
	 * \details The rules are in form of A -> Ba | Cb | ... | a | b | ..., where A, B, C ... \in N and a, b ... \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide a set of right hand sides of the rule
	 */
	void addRules ( SymbolType leftHandSide, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > rightHandSide );

	/**
	 * Get rules of the grammar.
	 *
	 * \returns rules of the grammar
	 */
	const ext::map < SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > & getRules ( ) const &;

	/**
	 * Get rules of the grammar.
	 *
	 * \returns rules of the grammar
	 */
	ext::map < SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > && getRules ( ) &&;

	/**
	 * Remove a rule of a grammar in form of A -> Ba or A -> a, where A, B \in N and a \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed removed, false othervise
	 */
	bool removeRule ( const SymbolType & leftHandSide, const ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > & rightHandSide );

	/**
	 * Remove a rule of a grammar in form of A -> a, where A \in N and a \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed removed, false othervise
	 */
	bool removeRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide );

	/**
	 * Remove a rule of a grammar in form of A -> Ba, where A, B \in N and a \in T*.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed removed, false othervise
	 */
	bool removeRule ( const SymbolType & leftHandSide, const ext::pair < ext::vector < SymbolType >, SymbolType > & rightHandSide );

	/**
	 * Add a new rule of a grammar in form of A -> Ba or A -> a, where A, B \in N and a \in T* as stored in combination of leftHandSide and rightHandSide represented as vector of symbols.
	 *
	 * \param leftHandSide the left hand side of the rule
	 * \param rightHandSide the right hand side of the rule
	 *
	 * \returns true if the rule was indeed added, false othervise
	 */
	bool addRawRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide );

	/**
	 * Get rules in most common format of rules as mapping from leftHandSide to rightHandSides represented as vectors of symbols.
	 *
	 * \returns rules of the grammar
	 */
	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > getRawRules ( ) const;

	/**
	 * Remove a rule of a grammar in form of A -> Ba or A -> a, where A, B \in N and a \in T* as stored in combination of leftHandSide and rightHandSide
	 *
	 * \returns true if the rule was indeed removed, false othervise
	 */
	bool removeRawRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide );

	/**
	 * Getter of initial symbol.
	 *
	 * \returns the initial symbol of the grammar
	 */
	const SymbolType & getInitialSymbol ( ) const & {
		return this->template accessComponent < InitialSymbol > ( ).get ( );
	}

	/**
	 * Getter of initial symbol.
	 *
	 * \returns the initial symbol of the grammar
	 */
	SymbolType && getInitialSymbol ( ) && {
		return std::move ( this->template accessComponent < InitialSymbol > ( ).get ( ) );
	}

	/**
	 * Setter of initial symbol.
	 *
	 * \param symbol new initial symbol of the grammar
	 *
	 * \returns true if the initial symbol was indeed changed
	 */
	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	/**
	 * Getter of nonterminal alphabet.
	 *
	 * \returns the nonterminal alphabet of the grammar
	 */
	const ext::set < SymbolType > & getNonterminalAlphabet ( ) const & {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	/**
	 * Getter of nonterminal alphabet.
	 *
	 * \returns the nonterminal alphabet of the grammar
	 */
	ext::set < SymbolType > && getNonterminalAlphabet ( ) && {
		return std::move ( this->template accessComponent < NonterminalAlphabet > ( ).get ( ) );
	}

	/**
	 * Adder of nonterminal symbol.
	 *
	 * \param symbol the new symbol to be added to nonterminal alphabet
	 *
	 * \returns true if the symbol was indeed added
	 */
	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	/**
	 * Setter of nonterminal alphabet.
	 *
	 * \param symbols completely new nonterminal alphabet
	 */
	void setNonterminalAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	/**
	 * Getter of terminal alphabet.
	 *
	 * \returns the terminal alphabet of the grammar
	 */
	const ext::set < SymbolType > & getTerminalAlphabet ( ) const & {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	/**
	 * Getter of terminal alphabet.
	 *
	 * \returns the terminal alphabet of the grammar
	 */
	ext::set < SymbolType > && getTerminalAlphabet ( ) && {
		return std::move ( this->template accessComponent < TerminalAlphabet > ( ).get ( ) );
	}

	/**
	 * Adder of terminal symbol.
	 *
	 * \param symbol the new symbol tuo be added to nonterminal alphabet
	 *
	 * \returns true if the symbol was indeed added
	 */
	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	/**
	 * Setter of terminal alphabet.
	 *
	 * \param symbol completely new nontemrinal alphabet
	 */
	void setTerminalAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	/**
	 * @copydoc alib::CommonBase<ObjectBase>::compare ( const ObjectBase & )
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * Actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns actual relation between two by type same grammar instances
	 */
	int compare ( const RightLG & other ) const;

	/**
	 * @copydoc alib::CommonBase<ObjectBase>::operator >> ( std::ostream & )
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc alib::CommonBase<ObjectBase>::operator std::string ( )
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * \brief The XML tag name of class.
	 *
	 * \details Intentionaly a static member function to be safe in the initialisation before the main function starts.
	 *
	 * \returns string representing the XML tag name of the class
	 */
	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RightLG";

		return xmlTagName;
	}

	/**
	 * Parsing from a sequence of xml tokens helper.
	 *
	 * \params input the iterator to sequence of xml tokens to parse from
	 *
	 * \returns the new instance of the grammar
	 */
	static RightLG parse ( ext::deque < sax::Token >::iterator & input );

	/**
	 * Helper for parsing of individual rules of the grammar from a sequence of xml tokens.
	 *
	 * \params input the iterator to sequence of xml tokens to parse from
	 * \params grammar the grammar to add the rule to
	 */
	static void parseRule ( ext::deque < sax::Token >::iterator & input, RightLG & grammar );

	/**
	 * Composing to a sequence of xml tokens helper.
	 *
	 * \param out sink for new xml tokens representing the grammar
	 * \param grammar the grammar to compose
	 */
	static void compose ( ext::deque < sax::Token > & out, const RightLG & grammar );

	/**
	 * Helper for composing rules of the grammar to a sequence of xml tokens.
	 *
	 * \param out sink for xml tokens representing the rules of the grammar
	 * \param grammar the grammar to compose
	 */
	static void composeRules ( ext::deque < sax::Token > & out, const RightLG & grammar );

	/**
	 * @copydoc alib::GrammarBase::inc()
	 */
	virtual object::ObjectBase * inc ( ) && override;

	/**
	 * Type of normalized grammar.
	 */
	typedef RightLG < > normalized_type;
};

template < class SymbolType >
RightLG < SymbolType >::RightLG ( SymbolType initialSymbol ) : RightLG ( ext::set < SymbolType > { initialSymbol }, ext::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
RightLG < SymbolType >::RightLG ( ext::set < SymbolType > nonterminalAlphabet, ext::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : core::Components < RightLG, ext::set < SymbolType >, component::Set, std::tuple < TerminalAlphabet, NonterminalAlphabet >, SymbolType, component::Value, InitialSymbol > ( std::move ( terminalAlphabet), std::move ( nonterminalAlphabet ), std::move ( initialSymbol ) ) {
}

template < class SymbolType >
GrammarBase * RightLG < SymbolType >::clone ( ) const {
	return new RightLG ( * this );
}

template < class SymbolType >
GrammarBase * RightLG < SymbolType >::plunder ( ) && {
	return new RightLG ( std::move ( * this ) );
}

template < class SymbolType >
bool RightLG < SymbolType >::addRule ( SymbolType leftHandSide, ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( rightHandSide.template is < ext::vector < SymbolType > > ( ) ) {
		for ( const auto & symbol : rightHandSide.template get < ext::vector < SymbolType > > ( ) )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ext::to_string ( symbol ) + " is not a terminal symbol" );
	} else {
		const ext::pair < ext::vector < SymbolType >, SymbolType > & rhs = rightHandSide.template get < ext::pair < ext::vector < SymbolType >, SymbolType > > ( );

		if ( !getNonterminalAlphabet ( ).count ( rhs.second ) )
			throw GrammarException ( "Symbol " + ext::to_string ( rhs.second ) + " is not a nonterminal symbol" );

		for ( const auto & symbol : rhs.first )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ext::to_string ( symbol ) + " is not a terminal symbol" );
	}

	return rules [ std::move ( leftHandSide ) ].insert ( std::move ( rightHandSide ) ).second;
}

template < class SymbolType >
bool RightLG < SymbolType >::addRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide ) {
	ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
bool RightLG < SymbolType >::addRule ( SymbolType leftHandSide, ext::pair < ext::vector < SymbolType >, SymbolType > rightHandSide ) {
	ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
void RightLG < SymbolType >::addRules ( SymbolType leftHandSide, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > & element : rightHandSide ) {
		if ( element.template is < ext::vector < SymbolType > > ( ) ) {
			for ( const auto & symbol : element.template get < ext::vector < SymbolType > > ( ) )
				if ( !getTerminalAlphabet ( ).count ( symbol ) )
					throw GrammarException ( "Symbol " + ext::to_string ( symbol ) + " is not a terminal symbol" );
		} else {
			const ext::pair < ext::vector < SymbolType >, SymbolType > & rhs = element.template get < ext::pair < ext::vector < SymbolType >, SymbolType > > ( );

			if ( !getNonterminalAlphabet ( ).count ( rhs.second ) )
				throw GrammarException ( "Symbol " + ext::to_string ( rhs.second ) + " is not a nonterminal symbol" );

			for ( const auto & symbol : rhs.first )
				if ( !getTerminalAlphabet ( ).count ( symbol ) )
					throw GrammarException ( "Symbol " + ext::to_string ( symbol ) + " is not a terminal symbol" );
		}
	}

	rules [ std::move ( leftHandSide ) ].insert ( ext::make_moveable_set ( rightHandSide ).begin ( ), ext::make_moveable_set ( rightHandSide ).end ( ) );
}

template < class SymbolType >
const ext::map < SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > & RightLG < SymbolType >::getRules ( ) const & {
	return rules;
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > && RightLG < SymbolType >::getRules ( ) && {
	return std::move ( rules );
}

template < class SymbolType >
bool RightLG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
bool RightLG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide ) {
	ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool RightLG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const ext::pair < ext::vector < SymbolType >, SymbolType > & rightHandSide ) {
	ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool RightLG < SymbolType >::addRawRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
	else if ( getNonterminalAlphabet ( ).count ( rightHandSide[rightHandSide.size ( ) - 1] ) )
		return addRule ( std::move ( leftHandSide ), ext::make_pair ( ext::vector < SymbolType > ( std::make_move_iterator ( rightHandSide.begin ( ) ), std::make_move_iterator ( rightHandSide.end ( ) ) - 1 ), std::move ( rightHandSide[rightHandSide.size ( ) - 1] ) ) );
	else
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > RightLG < SymbolType >::getRawRules ( ) const {
	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			if ( rhs.template is < ext::vector < SymbolType > > ( ) ) {
				res[rule.first].insert ( rhs.template get < ext::vector < SymbolType > > ( ) );
			} else {
				const auto & rhsTuple = rhs.template get < ext::pair < ext::vector < SymbolType >, SymbolType > > ( );
				ext::vector < SymbolType > tmp { rhsTuple.first.begin ( ), rhsTuple.first.end ( ) };
				tmp.push_back ( rhsTuple.second );
				res[rule.first].insert ( std::move ( tmp ) );
			}
		}

	return res;
}

template < class SymbolType >
bool RightLG < SymbolType >::removeRawRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		return removeRule ( leftHandSide, rightHandSide );
	else if ( getNonterminalAlphabet ( ).count ( rightHandSide[rightHandSide.size ( ) - 1] ) )
		return removeRule ( leftHandSide, ext::make_pair ( ext::vector < SymbolType > ( rightHandSide.begin ( ), rightHandSide.end ( ) - 1 ), rightHandSide[rightHandSide.size ( ) - 1] ) );
	else
		return removeRule ( leftHandSide, rightHandSide );
}

template < class SymbolType >
int RightLG < SymbolType >::compare ( const RightLG & other ) const {
	auto first = ext::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = ext::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void RightLG < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(LeftLG " << "nonterminalAlphabet = "
	    << getNonterminalAlphabet ( ) << "terminalAlphabet = "
	    << getTerminalAlphabet ( ) << "initialSymbol = "
	    << getInitialSymbol ( ) << "rules = "
	    << rules << ")";
}

template < class SymbolType >
RightLG < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
RightLG < SymbolType > RightLG < SymbolType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RightLG::getXmlTagName() );

	ext::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	ext::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	RightLG < SymbolType > grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RightLG::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void RightLG < SymbolType >::parseRule ( ext::deque < sax::Token >::iterator & input, RightLG & grammar ) {
	SymbolType lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS < SymbolType > ( input );
	ext::vector < SymbolType > rhs = GrammarFromXMLParser::parseRuleRHS < SymbolType > ( input );

	grammar.addRawRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void RightLG < SymbolType >::compose ( ext::deque < sax::Token > & out, const RightLG & grammar ) {
	out.emplace_back ( RightLG::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, grammar.getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, grammar.getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, grammar.getInitialSymbol ( ) );
	composeRules ( out, grammar );

	out.emplace_back ( RightLG::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void RightLG < SymbolType >::composeRules ( ext::deque < sax::Token > & out, const RightLG & grammar ) {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : grammar.getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRightLGRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
object::ObjectBase* RightLG < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace core {

/**
 * Helper class specifying constraints for the grammar's internal terminal alphabet component.
 *
 * \tparam SymbolType used for the terminal alphabet of the grammar.
 */
template < class SymbolType >
class SetConstraint< grammar::RightLG < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	/**
	 * Returns true if the terminal symbol is still used in some rule of the grammar.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \returns true if the symbol is used, false othervise
	 */
	static bool used ( const grammar::RightLG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > & rule : grammar.getRules ( ) ) {
			for ( const ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > & rhsTmp : rule.second )
				if ( rhsTmp.template is < ext::vector < SymbolType > > ( ) ) {
					const ext::vector < SymbolType > & rhs = rhsTmp.template get < ext::vector < SymbolType > > ( );

					if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
						return true;
				} else {
					const ext::vector < SymbolType > & rhs = rhsTmp.template get < ext::pair < ext::vector < SymbolType >, SymbolType > > ( ).first;

					if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
						return true;
				}

		}

		return false;
	}

	/**
	 * Returns true as all terminal symbols are possibly available to be terminal symbols.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \returns true
	 */
	static bool available ( const grammar::RightLG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	/**
	 * Throws runtime exception if the symbol requested to be terminal symbol is already in nonterminal alphabet.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \throws grammar::GrammarException of the tested symbol is in nonterminal alphabet
	 */
	static void valid ( const grammar::RightLG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

/**
 * Helper class specifying constraints for the grammar's internal nonterminal alphabet component.
 *
 * \tparam SymbolType used for the nonterminal alphabet of the grammar.
 */
template < class SymbolType >
class SetConstraint< grammar::RightLG < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	/**
	 * Returns true if the nonterminal symbol is still used in some rule of the grammar or if it is the initial symbol of the grammar.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \returns true if the symbol is used, false othervise
	 */
	static bool used ( const grammar::RightLG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > & rule : grammar.getRules ( ) ) {
			if ( rule.first == symbol )
				return true;

			for ( const ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > & rhsTmp : rule.second )
				if ( rhsTmp.template is < ext::pair < ext::vector < SymbolType >, SymbolType > > ( ) ) {
					const ext::pair < ext::vector < SymbolType >, SymbolType > & rhs = rhsTmp.template get < ext::pair < ext::vector < SymbolType >, SymbolType > > ( );

					if ( rhs.second == symbol )
						return true;
				}

		}

		if ( grammar.template accessComponent < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	/**
	 * Returns true as all terminal symbols are possibly available to be nonterminal symbols.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \returns true
	 */
	static bool available ( const grammar::RightLG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	/**
	 * Throws runtime exception if the symbol requested to be nonterminal symbol is already in terminal alphabet.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \throws grammar::GrammarException of the tested symbol is in nonterminal alphabet
	 */
	static void valid ( const grammar::RightLG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

/**
 * Helper class specifying constraints for the grammar's internal initial symbol element.
 *
 * \tparam SymbolType used for the initial symbol of the grammar.
 */
template < class SymbolType >
class ElementConstraint< grammar::RightLG < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	/**
	 * Returns true if the symbol requested to be initial is available in nonterminal alphabet.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 *
	 * \returns true if the tested symbol is in nonterminal alphabet
	 */
	static bool available ( const grammar::RightLG < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	/**
	 * All symbols are valid as initial symbols.
	 *
	 * \param grammar the tested grammar
	 * \param symbol the tested symbol
	 */
	static void valid ( const grammar::RightLG < SymbolType > &, const SymbolType & ) {
	}
};

/**
 * Helper for normalisation of types specified by templates used as internal datatypes of symbols.
 *
 * \returns new instance of the grammar with default template parameters or unmodified instance if the template parameters were already default ones
 */
template < class SymbolType >
struct normalize < grammar::RightLG < SymbolType >, typename std::enable_if < ! std::is_same < grammar::RightLG < SymbolType >, grammar::RightLG < > >::value >::type > {
	static grammar::RightLG < > eval ( grammar::RightLG < SymbolType > && value ) {
		ext::set < DefaultSymbolType > nonterminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getNonterminalAlphabet ( ) );
		ext::set < DefaultSymbolType > terminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getTerminalAlphabet ( ) );
		DefaultSymbolType initialSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( value ).getInitialSymbol ( ) );

		grammar::RightLG < > res ( std::move ( nonterminals ), std::move ( terminals ), std::move ( initialSymbol ) );

		for ( std::pair < SymbolType, ext::set < ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > > > && rule : ext::make_moveable_map ( std::move ( value ).getRules ( ) ) ) {

			ext::set < ext::variant < ext::vector < DefaultSymbolType >, ext::pair < ext::vector < DefaultSymbolType >, DefaultSymbolType > > > rhs;
			for ( ext::variant < ext::vector < SymbolType >, ext::pair < ext::vector < SymbolType >, SymbolType > > && target : ext::make_moveable_set ( rule.second ) )
				rhs.insert ( grammar::GrammarNormalize::normalizeRHS ( std::move ( target ) ) );

			DefaultSymbolType lhs = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( rule.first ) );

			res.addRules ( std::move ( lhs ), std::move ( rhs ) );
		}

		return res;
	}
};

} /* namespace core */

#endif /* RIGHT_LG_H_ */
