/*
 * NonlinearFullAndLinearIndex.h
 *
 *  Created on: Apr 13, 2017
 *      Author: Jan Travnicek
 */

#ifndef ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_
#define ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_

#include <alib/string>
#include <sstream>

#include <alphabet/RankedSymbol.h>
#include <alphabet/xml/RankedSymbol.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsVector.h>
#include <container/xml/ObjectsVector.h>

#include <primitive/xml/Integer.h>

#include <indexes/stringology/PositionHeap.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace arbology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType, class RankType = DefaultRankType, template < typename > class StringIndex = indexes::stringology::PositionHeap >
class NonlinearFullAndLinearIndex final : public object::ObjectBase {
protected:
	StringIndex < common::ranked_symbol < SymbolType, RankType > > m_StringIndex;
	ext::vector < int > m_JumpTable;
	ext::vector < unsigned > m_Repeats;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit NonlinearFullAndLinearIndex ( StringIndex < common::ranked_symbol < SymbolType, RankType > > stringIndex, ext::vector < int > jumpTable, ext::vector < unsigned > repeats );

	/**
	 * @return Root node of the trie
	 */
	const StringIndex < common::ranked_symbol < SymbolType, RankType > > & getStringIndex ( ) const &;

	StringIndex < common::ranked_symbol < SymbolType, RankType > > && getStringIndex ( ) &&;

	/**
	 * @return subtree jump table
	 */
	const ext::vector < int > & getJumps ( ) const &;

	ext::vector < int > && getJumps ( ) &&;

	/**
	 * @return repeats
	 */
	const ext::vector < unsigned > & getRepeats ( ) const &;

	ext::vector < unsigned > && getRepeats ( ) &&;

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return m_StringIndex.getAlphabet ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( m_StringIndex.getAlphabet ( ) );
	}

	/**
	 * Sets the compressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setStringIndex ( StringIndex < common::ranked_symbol < SymbolType, RankType > > stringIndex );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return m_StringIndex.removeSymbolFromAlphabet ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NonlinearFullAndLinearIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual object::ObjectBase * inc ( ) &&;

	typedef NonlinearFullAndLinearIndex < > normalized_type;
};

} /* namespace arbology */

} /* namespace indexes */

namespace indexes {

namespace arbology {

template < class SymbolType, class RankType, template < typename > class StringIndex >
NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::NonlinearFullAndLinearIndex ( StringIndex < common::ranked_symbol < SymbolType, RankType > > stringIndex, ext::vector < int > jumpTable, ext::vector < unsigned > repeats ) : m_StringIndex ( std::move ( stringIndex ) ), m_JumpTable ( std::move ( jumpTable ) ), m_Repeats ( std::move ( repeats ) ) {
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
object::ObjectBase * NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::clone ( ) const {
	return new NonlinearFullAndLinearIndex ( * this );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
object::ObjectBase * NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::plunder ( ) && {
	return new NonlinearFullAndLinearIndex ( std::move ( * this ) );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
const StringIndex < common::ranked_symbol < SymbolType, RankType > > & NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::getStringIndex ( ) const & {
	return m_StringIndex;
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
StringIndex < common::ranked_symbol < SymbolType, RankType > > && NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::getStringIndex ( ) && {
	return std::move ( m_StringIndex );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
const ext::vector < int > & NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::getJumps ( ) const & {
	return m_JumpTable;
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
ext::vector < int > && NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::getJumps ( ) && {
	return std::move ( m_JumpTable );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
const ext::vector < unsigned > & NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::getRepeats ( ) const & {
	return m_Repeats;
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
ext::vector < unsigned > && NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::getRepeats ( ) && {
	return std::move ( m_Repeats );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
void NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::setStringIndex ( StringIndex < common::ranked_symbol < SymbolType, RankType > > stringIndex ) {
	this->m_StringIndex = std::move ( stringIndex );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
void NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::operator >>( std::ostream & out ) const {
	out << "(NonlinearFullAndLinearIndex " << this->m_StringIndex << ", " << m_JumpTable << ")";
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
int NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::compare ( const NonlinearFullAndLinearIndex & other ) const {
	auto first = ext::tie ( getStringIndex ( ), getJumps ( ), getRepeats ( ) );
	auto second = ext::tie ( other.getStringIndex ( ), other.getJumps ( ), other.getRepeats ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
object::ObjectBase * NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >::inc ( ) && {
	return new object::UniqueObject ( object::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace arbology */

} /* namespace indexes */

namespace core {

template < class SymbolType, class RankType, template < typename > class StringIndex >
struct normalize < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >, typename std::enable_if < ! std::is_same < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex >, indexes::arbology::NonlinearFullAndLinearIndex < > >::value >::type > {
	static indexes::arbology::NonlinearFullAndLinearIndex < > eval ( indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > && value ) {
		StringIndex < common::ranked_symbol < > > stringIndex = core::normalize < StringIndex < common::ranked_symbol < SymbolType, RankType > > >::eval ( std::move ( value ).getStringIndex ( ) );

		return indexes::arbology::NonlinearFullAndLinearIndex < > ( std::move ( stringIndex ), std::move ( value ).getJumps ( ), std::move ( value ).getRepeats ( ) );
	}
};

template < class SymbolType, class RankType, template < typename > class StringIndex >
struct xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > > {
	static indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > & data );
};

template < class SymbolType, class RankType, template < typename > class StringIndex >
indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	StringIndex < common::ranked_symbol < SymbolType, RankType > > stringIndex = core::xmlApi < StringIndex < common::ranked_symbol < SymbolType, RankType > > >::parse ( input );
	ext::vector < int > jumps = core::xmlApi < ext::vector < int > >::parse ( input );
	ext::vector < unsigned > repeats = core::xmlApi < ext::vector < unsigned > >::parse ( input );

	indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > res ( std::move ( stringIndex ), std::move ( jumps ), std::move ( repeats ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
bool xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
const std::string & xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > >::xmlTagName ( ) {
	static std::string xmlTagName = "NonlinearFullAndLinearIndex";

	return xmlTagName;
}

template < class SymbolType, class RankType, template < typename > class StringIndex >
void xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > >::compose ( ext::deque < sax::Token > & output, const indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType, StringIndex > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	core::xmlApi < StringIndex < common::ranked_symbol < SymbolType, RankType > > >::compose ( output, index.getStringIndex ( ) );
	core::xmlApi < ext::vector < int > >::compose ( output, index.getJumps ( ) );
	core::xmlApi < ext::vector < unsigned > >::compose ( output, index.getRepeats ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace core */

#endif /* ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_ */
