/*
 * PositionHeap.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef POSITION_HEAP_H_
#define POSITION_HEAP_H_

#include <alib/string>
#include <alib/set>
#include <alib/trie>
#include <alib/iostream>
#include <alib/algorithm>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/Object.h>
#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsTrie.h>
#include <container/ObjectsVector.h>

#include <container/xml/ObjectsSet.h>
#include <container/xml/ObjectsTrie.h>
#include <container/xml/ObjectsVector.h>

#include <primitive/Unsigned.h>
#include <primitive/xml/Unsigned.h>

#include <alphabet/common/SymbolNormalize.h>
#include <indexes/common/IndexesNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class PositionHeap final : public object::ObjectBase, public core::Components < PositionHeap < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
protected:
	ext::trie < SymbolType, unsigned > m_trie;
	ext::vector < SymbolType > m_string;

public:
	/**
	 * @copydoc PositionHeap::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc PositionHeap::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit PositionHeap ( ext::set < SymbolType > edgeAlphabet, ext::trie < SymbolType, unsigned > trie, ext::vector < SymbolType > string );
	explicit PositionHeap ( ext::trie < SymbolType, unsigned > trie, ext::vector < SymbolType > string );

	void checkTrie ( const ext::trie < SymbolType, unsigned > & trie );

	/**
	 * @return Root node of the trie
	 */
	const ext::trie < SymbolType, unsigned > & getRoot ( ) const &;

	ext::trie < SymbolType, unsigned > && getRoot ( ) &&;

	const ext::vector < SymbolType > & getString ( ) const &;

	ext::vector < SymbolType > && getString ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( ext::trie < SymbolType, unsigned > tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromEdgeAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PositionHeap & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual object::ObjectBase * inc ( ) &&;

	typedef PositionHeap < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
PositionHeap < SymbolType >::PositionHeap ( ext::set < SymbolType > edgeAlphabet, ext::trie < SymbolType, unsigned > trie, ext::vector < SymbolType > string ) : core::Components < PositionHeap, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( edgeAlphabet ) ), m_trie ( std::move ( trie ) ), m_string ( std::move ( string ) ) {
	checkTrie ( this->m_trie );
	// TODO check validity of the string like in LinearString
}

template < class SymbolType >
PositionHeap < SymbolType >::PositionHeap ( ext::trie < SymbolType, unsigned > trie, ext::vector < SymbolType > string ) : PositionHeap ( computeMinimalEdgeAlphabet ( trie ), trie, std::move ( string ) ) {
}

template < class SymbolType >
object::ObjectBase * PositionHeap < SymbolType >::clone ( ) const {
	return new PositionHeap ( * this );
}

template < class SymbolType >
object::ObjectBase * PositionHeap < SymbolType >::plunder ( ) && {
	return new PositionHeap ( std::move ( * this ) );
}

template < class SymbolType >
void PositionHeap < SymbolType >::checkTrie ( const ext::trie < SymbolType, unsigned > & trie ) {
	for ( const std::pair < const SymbolType, ext::trie < SymbolType, unsigned > > & child : trie.getChildren ( ) ) {
		if ( ! getAlphabet ( ).count ( child.first ) )
			throw exception::CommonException ( "Symbol " + ext::to_string ( child.first ) + "not in the alphabet." );
		checkTrie ( child.second );
	}
}

template < class SymbolType >
const ext::trie < SymbolType, unsigned > & PositionHeap < SymbolType >::getRoot ( ) const & {
	return m_trie;
}

template < class SymbolType >
ext::trie < SymbolType, unsigned > && PositionHeap < SymbolType >::getRoot ( ) && {
	return std::move ( m_trie );
}

template < class SymbolType >
const ext::vector < SymbolType > & PositionHeap < SymbolType >::getString ( ) const & {
	return m_string;
}

template < class SymbolType >
ext::vector < SymbolType > && PositionHeap < SymbolType >::getString ( ) && {
	return std::move ( m_string );
}

template < class SymbolType >
void PositionHeap < SymbolType >::setTree ( ext::trie < SymbolType, unsigned > trie ) {
	checkTrie ( trie );
	this->m_trie = std::move ( trie ).plunder ( );
}

template < class SymbolType >
void PositionHeap < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(PositionHeap " << this->m_trie << ")";
}

template < class SymbolType >
int PositionHeap < SymbolType >::compare ( const PositionHeap & other ) const {
	auto first = ext::tie ( getRoot ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.getRoot ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
PositionHeap < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
object::ObjectBase* PositionHeap < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace core {

template < class SymbolType >
class SetConstraint < indexes::stringology::PositionHeap < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {

	static bool used ( const ext::trie < SymbolType, unsigned > & trie, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::trie < SymbolType, unsigned > > & child : trie.getChildren ( ) ) {
			if ( symbol == child.first || checkTrie ( trie, child.second ) )
				return true;
		}
		return false;
	}

public:
	static bool used ( const indexes::stringology::PositionHeap < SymbolType > & index, const SymbolType & symbol ) {
		const ext::vector < SymbolType > & content = index.getString ( );
		return std::find ( content.begin(), content.end(), symbol ) != content.end() || used ( index.getRoot ( ), symbol );
	}

	static bool available ( const indexes::stringology::PositionHeap < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::PositionHeap < SymbolType > &, const SymbolType & ) {
	}
};

template < >
struct normalize < indexes::stringology::PositionHeap < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > > {
	static indexes::stringology::PositionHeap < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > && eval ( indexes::stringology::PositionHeap < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > && source ) {
		return std::move ( source );
	}
};

template < class SymbolType, class RankType >
struct normalize < indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > > {
	static indexes::stringology::PositionHeap < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > eval ( indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > && source ) {
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( source ).getAlphabet ( ) );
		ext::trie < common::ranked_symbol < DefaultSymbolType, DefaultRankType >, unsigned > trie = indexes::IndexesNormalize::normalizeRankedTrie ( std::move ( source ).getRoot ( ) );
		ext::vector < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > string = alphabet::SymbolNormalize::normalizeRankedSymbols ( std::move ( source ).getString ( ) );

		return indexes::stringology::PositionHeap < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > ( std::move ( alphabet ), std::move ( trie ), std::move ( string ) );
	}
};

template < class SymbolType >
struct normalize < indexes::stringology::PositionHeap < SymbolType >, typename std::enable_if < ! std::is_same < indexes::stringology::PositionHeap < SymbolType >, indexes::stringology::PositionHeap < > >::value >::type > {
	static indexes::stringology::PositionHeap < > eval ( indexes::stringology::PositionHeap < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::trie < DefaultSymbolType, unsigned > trie = indexes::IndexesNormalize::normalizeTrie ( std::move ( value ).getRoot ( ) );
		ext::vector < DefaultSymbolType > string = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( value ).getString ( ) );

		return indexes::stringology::PositionHeap < > ( std::move ( alphabet ), std::move ( trie ), std::move ( string ) );
	}
};

template < class SymbolType >
struct xmlApi < indexes::stringology::PositionHeap < SymbolType > > {
	static indexes::stringology::PositionHeap < SymbolType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::PositionHeap < SymbolType > & data );
};

template < class SymbolType >
indexes::stringology::PositionHeap < SymbolType > xmlApi < indexes::stringology::PositionHeap < SymbolType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	ext::set < SymbolType > edgeAlphabet = core::xmlApi < ext::set < SymbolType > >::parse ( input );
	ext::trie < SymbolType, unsigned > root = core::xmlApi < ext::trie < SymbolType, unsigned > >::parse ( input );
	ext::vector < SymbolType > string = core::xmlApi < ext::vector < SymbolType > >::parse ( input );
	indexes::stringology::PositionHeap < SymbolType > res ( std::move ( edgeAlphabet ), std::move ( root ), std::move ( string ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType >
bool xmlApi < indexes::stringology::PositionHeap < SymbolType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType >
const std::string & xmlApi < indexes::stringology::PositionHeap < SymbolType > >::xmlTagName ( ) {
	static std::string xmlTagName = "PositionHeap";

	return xmlTagName;
}

template < class SymbolType >
void xmlApi < indexes::stringology::PositionHeap < SymbolType > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::PositionHeap < SymbolType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	core::xmlApi < ext::set < SymbolType > >::compose ( output, index.getAlphabet ( ) );
	core::xmlApi < ext::trie < SymbolType, unsigned > >::compose ( output, index.getRoot ( ) );
	core::xmlApi < ext::vector < SymbolType > >::compose ( output, index.getString ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace core */

#endif /* POSITION_HEAP_H_ */
