/*
 * FormalRegExpAlternation.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef FORMAL_REG_EXP_ALTERNATION_H_
#define FORMAL_REG_EXP_ALTERNATION_H_

#include <sstream>
#include <alib/utility>

#include <exception/CommonException.h>

#include "FormalRegExpElement.h"

namespace regexp {

/**
 * Represents alternation operator in the regular expression. Contains list of FormalRegExpElement
 * as operands of the operator.
 */
template < class SymbolType >
class FormalRegExpAlternation : public FormalRegExpElement < SymbolType >, public ext::BinaryNode < ext::smart_ptr < FormalRegExpElement < SymbolType > >, ext::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpAlternation < SymbolType > > {
public:
	void accept ( typename FormalRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit FormalRegExpAlternation ( FormalRegExpElement < SymbolType > && left, FormalRegExpElement < SymbolType > && right );

	explicit FormalRegExpAlternation ( const FormalRegExpElement < SymbolType > & left, const FormalRegExpElement < SymbolType > & right );

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	FormalRegExpAlternation < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	FormalRegExpAlternation < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	ext::smart_ptr < UnboundedRegExpElement < SymbolType > > asUnbounded ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc FormalRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( ext::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc FormalRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const ext::set < SymbolType > & alphabet ) const override;

	/**
	 * @return elements
	 */
	const FormalRegExpElement < SymbolType > & getLeftElement ( ) const;
	const FormalRegExpElement < SymbolType > & getRightElement ( ) const;

	/**
	 * @return elements
	 */
	FormalRegExpElement < SymbolType > & getLeftElement ( );
	FormalRegExpElement < SymbolType > & getRightElement ( );

	/**
	 * @param element to append
	 */
	void setLeftElement ( FormalRegExpElement < SymbolType > && element );
	void setLeftElement ( const FormalRegExpElement < SymbolType > & element );

	/**
	 * @param element to append
	 */
	void setRightElement ( FormalRegExpElement < SymbolType > && element );
	void setRightElement ( const FormalRegExpElement < SymbolType > & element );

	int compare ( const FormalRegExpElement < SymbolType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	int compare ( const FormalRegExpAlternation & ) const;

	/**
	 * @copydoc FormalRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	explicit operator std::string ( ) const override;

	virtual ext::smart_ptr < FormalRegExpElement < DefaultSymbolType > > normalize ( ) && override {
		return ext::smart_ptr < FormalRegExpElement < DefaultSymbolType > > ( new FormalRegExpAlternation < DefaultSymbolType > ( std::move ( * std::move ( getLeftElement ( ) ).normalize ( ) ), std::move ( * std::move ( getRightElement ( ) ).normalize ( ) ) ) );
	}
};

} /* namespace regexp */

#include "../unbounded/UnboundedRegExpAlternation.h"

namespace regexp {

template < class SymbolType >
FormalRegExpAlternation < SymbolType >::FormalRegExpAlternation ( FormalRegExpElement < SymbolType > && leftElement, FormalRegExpElement < SymbolType > && rightElement ) : ext::BinaryNode < ext::smart_ptr < FormalRegExpElement < SymbolType > >, ext::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpAlternation < SymbolType > > ( ext::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( leftElement ).plunder ( ) ), ext::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( rightElement ).plunder ( ) ) ) {
}

template < class SymbolType >
FormalRegExpAlternation < SymbolType >::FormalRegExpAlternation ( const FormalRegExpElement < SymbolType > & left, const FormalRegExpElement < SymbolType > & right ) : FormalRegExpAlternation ( ext::move_copy ( left ), ext::move_copy ( right ) ) {
}

template < class SymbolType >
const FormalRegExpElement < SymbolType > & FormalRegExpAlternation < SymbolType >::getLeftElement ( ) const {
	return * this->getLeft ( );
}

template < class SymbolType >
const FormalRegExpElement < SymbolType > & FormalRegExpAlternation < SymbolType >::getRightElement ( ) const {
	return * this->getRight ( );
}

template < class SymbolType >
FormalRegExpElement < SymbolType > & FormalRegExpAlternation < SymbolType >::getLeftElement ( ) {
	return * this->getLeft ( );
}

template < class SymbolType >
FormalRegExpElement < SymbolType > & FormalRegExpAlternation < SymbolType >::getRightElement ( ) {
	return * this->getRight ( );
}

template < class SymbolType >
void FormalRegExpAlternation < SymbolType >::setLeftElement ( FormalRegExpElement < SymbolType > && element ) {
	this->setLeft ( ext::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType >
void FormalRegExpAlternation < SymbolType >::setLeftElement ( const FormalRegExpElement < SymbolType > & element ) {
	setLeftElement ( ext::move_copy ( element ) );
}

template < class SymbolType >
void FormalRegExpAlternation < SymbolType >::setRightElement ( FormalRegExpElement < SymbolType > && element ) {
	this->setRight ( ext::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType >
void FormalRegExpAlternation < SymbolType >::setRightElement ( const FormalRegExpElement < SymbolType > & element ) {
	setRightElement ( ext::move_copy ( element ) );
}

template < class SymbolType >
FormalRegExpAlternation < SymbolType > * FormalRegExpAlternation < SymbolType >::clone ( ) const {
	return new FormalRegExpAlternation ( * this );
}

template < class SymbolType >
FormalRegExpAlternation < SymbolType > * FormalRegExpAlternation < SymbolType >::plunder ( ) && {
	return new FormalRegExpAlternation ( std::move ( * this ) );
}

template < class SymbolType >
ext::smart_ptr < UnboundedRegExpElement < SymbolType > > FormalRegExpAlternation < SymbolType >::asUnbounded ( ) const {
	UnboundedRegExpAlternation < SymbolType > * res = new UnboundedRegExpAlternation < SymbolType > ( );

	res->appendElement ( std::move ( * getLeftElement ( ).asUnbounded ( ) ) );
	res->appendElement ( std::move ( * getRightElement ( ).asUnbounded ( ) ) );

	return ext::smart_ptr < UnboundedRegExpElement < SymbolType > > ( res );
}

template < class SymbolType >
int FormalRegExpAlternation < SymbolType >::compare ( const FormalRegExpAlternation < SymbolType > & other ) const {
	int res = getLeftElement ( ).compare ( other.getLeftElement ( ) );

	if ( res == 0 ) res = getRightElement ( ).compare ( other.getRightElement ( ) );

	return res;
}

template < class SymbolType >
void FormalRegExpAlternation < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(FormalRegExpAlternation";
	out << " " << getLeftElement ( );
	out << " " << getRightElement ( );
	out << ")";
}

template < class SymbolType >
bool FormalRegExpAlternation < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	if ( getLeftElement ( ).testSymbol ( symbol ) ) return true;

	if ( getRightElement ( ).testSymbol ( symbol ) ) return true;

	return false;
}

template < class SymbolType >
void FormalRegExpAlternation < SymbolType >::computeMinimalAlphabet ( ext::set < SymbolType > & alphabet ) const {
	getLeftElement ( ).computeMinimalAlphabet ( alphabet );
	getRightElement ( ).computeMinimalAlphabet ( alphabet );
}

template < class SymbolType >
bool FormalRegExpAlternation < SymbolType >::checkAlphabet ( const ext::set < SymbolType > & alphabet ) const {
	return getLeftElement ( ).checkAlphabet ( alphabet ) && getRightElement ( ).checkAlphabet ( alphabet );
}

template < class SymbolType >
FormalRegExpAlternation < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_ALTERNATION_H_ */
