/*
 * Epsilon.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_H_
#define EPSILON_H_

#include <alib/set>
#include <alib/vector>

#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "StringBase.h"
#include "StringFeatures.h"
#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace string {

class GeneralAlphabet;

/**
 * Represents epsilon. Regular expression is stored
 * as a tree of EpsilonElement.
 */
template < class SymbolType >
class Epsilon final : public StringBase, public core::Components < Epsilon < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
	static const ext::vector < SymbolType > content;

public:
	Epsilon ( );

	Epsilon ( ext::set < SymbolType > alphabet );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > & getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void extendAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbols ) );
	}

	/**
	 * @return List of symbols forming string (const version).
	 */
	const ext::vector < SymbolType > & getContent ( ) const;

	/**
	 * @return true
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const Epsilon & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static Epsilon EPSILON;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Epsilon";

		return xmlTagName;
	}

	virtual object::ObjectBase * inc ( ) &&;

	typedef Epsilon < > normalized_type;
};

template < class SymbolType >
Epsilon < SymbolType >::Epsilon ( ext::set < SymbolType > alphabet ) : core::Components < Epsilon, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( alphabet ) ) {
}

template < class SymbolType >
Epsilon < SymbolType >::Epsilon() : Epsilon ( ext::set< SymbolType > ( ) ) {
}

template < class SymbolType >
StringBase* Epsilon < SymbolType >::clone() const {
	return new Epsilon(*this);
}

template < class SymbolType >
StringBase* Epsilon < SymbolType >::plunder() && {
	return new Epsilon(std::move(*this));
}

template < class SymbolType >
const ext::vector<SymbolType> & Epsilon < SymbolType >::getContent() const {
	return Epsilon::content;
}

template < class SymbolType >
bool Epsilon < SymbolType >::isEmpty() const {
	return true;
}

template < class SymbolType >
int Epsilon < SymbolType >::compare(const Epsilon& other) const {
	static ext::compare<ext::set<SymbolType>> comp;
	return comp(getAlphabet(), other.getAlphabet());
}

template < class SymbolType >
void Epsilon < SymbolType >::operator>>(std::ostream& out) const {
	out << "(Epsilon)";
}

template < class SymbolType >
Epsilon < SymbolType >::operator std::string() const {
	return "E";
}

template < class SymbolType >
Epsilon < SymbolType > Epsilon < SymbolType >::EPSILON = Epsilon < SymbolType >();

template < class SymbolType >
const ext::vector<SymbolType> Epsilon < SymbolType >::content;

template < class SymbolType >
object::ObjectBase* Epsilon < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace string */

namespace core {

template < class SymbolType >
class SetConstraint< string::Epsilon < SymbolType >, SymbolType, string::GeneralAlphabet > {
public:
	static bool used ( const string::Epsilon < SymbolType > &, const SymbolType & ) {
		return false;
	}

	static bool available ( const string::Epsilon < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const string::Epsilon < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < string::Epsilon < SymbolType >, typename std::enable_if < ! std::is_same < string::Epsilon < SymbolType >, string::Epsilon < > >::value >::type > {
	static string::Epsilon < > eval ( string::Epsilon < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		return string::Epsilon < > ( std::move ( alphabet ) );
	}
};

} /* namespace core */

#endif /* EPSILON_H_ */
