/*
 * RankedPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef RANKED_PATTERN_H_
#define RANKED_PATTERN_H_

#include <alib/string>
#include <alib/set>
#include <alib/tree>
#include <alib/iostream>
#include <alib/algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <common/ranked_symbol.hpp>
#include <alphabet/xml/RankedSymbol.h>

#include "../TreeFeatures.h"
#include "../TreeException.h"
#include "../TreeBase.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include <tree/common/TreeNormalize.h>

#include <core/normalize.hpp>

namespace tree {

class GeneralAlphabet;
class SubtreeWildcard;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
template < class SymbolType, class RankType >
class RankedPattern final : public TreeBase, public core::Components < RankedPattern < SymbolType, RankType >, ext::set < common::ranked_symbol < SymbolType, RankType > >, component::Set, GeneralAlphabet, common::ranked_symbol < SymbolType, RankType >, component::Value, SubtreeWildcard > {
	ext::tree < common::ranked_symbol < SymbolType, RankType > > m_content;

	void checkAlphabet ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & pattern ) const;

	void checkArities ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & data ) const;

public:
	/**
	 * @copydoc ext::tree < common::ranked_symbol < SymbolType, RankType > >::clone() const
	 */
	virtual TreeBase * clone ( ) const;

	/**
	 * @copydoc ext::tree < common::ranked_symbol < SymbolType, RankType > >::plunder() const
	 */
	virtual TreeBase * plunder ( ) &&;

	explicit RankedPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::tree < common::ranked_symbol < SymbolType, RankType > > pattern );
	explicit RankedPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, ext::tree < common::ranked_symbol < SymbolType, RankType > > pattern );
	explicit RankedPattern ( const UnrankedPattern < SymbolType > & other );

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void extendAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const common::ranked_symbol < SymbolType, RankType > & getSubtreeWildcard ( ) const & {
		return this->template accessComponent < SubtreeWildcard > ( ).get ( );
	}

	common::ranked_symbol < SymbolType, RankType > && getSubtreeWildcard ( ) && {
		return std::move ( this->template accessComponent < SubtreeWildcard > ( ).get ( ) );
	}

	/**
	 * @return Root node of the regular expression pattern
	 */
	const ext::tree < common::ranked_symbol < SymbolType, RankType > > & getContent ( ) const &;

	ext::tree < common::ranked_symbol < SymbolType, RankType > > && getContent ( ) &&;

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( ext::tree < common::ranked_symbol < SymbolType, RankType > > pattern );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RankedPattern";

		return xmlTagName;
	}

	static RankedPattern parse ( ext::deque < sax::Token >::iterator & input );

	static void compose ( ext::deque < sax::Token > & out, const RankedPattern & pattern );

	virtual object::ObjectBase * inc ( ) &&;

	typedef RankedPattern < > normalized_type;
};

} /* namespace tree */

#include "../unranked/UnrankedPattern.h"

namespace tree {

template < class SymbolType, class RankType >
RankedPattern < SymbolType, RankType >::RankedPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::tree < common::ranked_symbol < SymbolType, RankType > > pattern ) : core::Components < RankedPattern, ext::set < common::ranked_symbol < SymbolType, RankType > >, component::Set, GeneralAlphabet, common::ranked_symbol < SymbolType, RankType >, component::Value, SubtreeWildcard > ( std::move ( alphabet ), std::move ( subtreeWildcard ) ), m_content ( std::move ( pattern ) ) {
	checkAlphabet ( m_content );
	checkArities ( m_content );
}

template < class SymbolType, class RankType >
RankedPattern < SymbolType, RankType >::RankedPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, ext::tree < common::ranked_symbol < SymbolType, RankType > > pattern ) : RankedPattern ( subtreeWildcard, ext::set < common::ranked_symbol < SymbolType, RankType > > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ) + ext::set < common::ranked_symbol < SymbolType, RankType > > { subtreeWildcard }, pattern ) {
}

template < class SymbolType, class RankType >
RankedPattern < SymbolType, RankType >::RankedPattern ( const UnrankedPattern < SymbolType > & other ) : RankedPattern ( common::ranked_symbol < SymbolType, RankType > ( other.getSubtreeWildcard ( ), 0 ), TreeAuxiliary::toRankedTree < SymbolType, RankType > ( other.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
TreeBase * RankedPattern < SymbolType, RankType >::clone ( ) const {
	return new RankedPattern ( * this );
}

template < class SymbolType, class RankType >
TreeBase * RankedPattern < SymbolType, RankType >::plunder ( ) && {
	return new RankedPattern ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const ext::tree < common::ranked_symbol < SymbolType, RankType > > & RankedPattern < SymbolType, RankType >::getContent ( ) const & {
	return m_content;
}

template < class SymbolType, class RankType >
ext::tree < common::ranked_symbol < SymbolType, RankType > > && RankedPattern < SymbolType, RankType >::getContent ( ) && {
	return std::move ( m_content );
}

template < class SymbolType, class RankType >
void RankedPattern < SymbolType, RankType >::checkAlphabet ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & data ) const {
	ext::set < common::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	ext::set < common::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType, class RankType >
void RankedPattern < SymbolType, RankType >::checkArities ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & data ) const {
	if ( ( size_t ) data.getData ( ).getRank ( ) != data.getChildren ( ).size ( ) )
		throw exception::CommonException ( "Invalid rank." );

	for ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & child : data )
		checkArities ( child );
}

template < class SymbolType, class RankType >
void RankedPattern < SymbolType, RankType >::setTree ( ext::tree < common::ranked_symbol < SymbolType, RankType > > pattern ) {
	checkAlphabet ( pattern );
	checkArities ( pattern );

	this->m_content = std::move ( pattern );
}

template < class SymbolType, class RankType >
void RankedPattern < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(RankedPattern " << this->m_content << ")";
}

template < class SymbolType, class RankType >
int RankedPattern < SymbolType, RankType >::compare ( const RankedPattern & other ) const {
	auto first = ext::tie ( m_content, getAlphabet(), getSubtreeWildcard() );
	auto second = ext::tie ( other.m_content, other.getAlphabet(), other.getSubtreeWildcard() );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void RankedPattern < SymbolType, RankType >::nicePrint ( std::ostream & os ) const {
	m_content.nicePrint ( os );
}

template < class SymbolType, class RankType >
RankedPattern < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType >
RankedPattern < SymbolType, RankType > RankedPattern < SymbolType, RankType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedPattern::getXmlTagName() );
	common::ranked_symbol < SymbolType, RankType > subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol < common::ranked_symbol < SymbolType, RankType > > ( input );
	ext::set < common::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < common::ranked_symbol < SymbolType, RankType > > ( input );
	ext::tree < common::ranked_symbol < SymbolType, RankType > > root = TreeFromXMLParser::parseTreeContent < common::ranked_symbol < SymbolType, RankType > > ( input );
	RankedPattern < SymbolType, RankType > tree ( std::move ( subtreeWildcardSymbol ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedPattern::getXmlTagName() );
	return tree;
}

template < class SymbolType, class RankType >
void RankedPattern < SymbolType, RankType >::compose ( ext::deque < sax::Token > & out, const RankedPattern & pattern ) {
	out.emplace_back ( RankedPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, pattern.getSubtreeWildcard() );
	TreeToXMLComposer::composeAlphabet ( out, pattern.getAlphabet() );
	TreeToXMLComposer::composeContent ( out, pattern.getContent ( ) );
	out.emplace_back ( RankedPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
object::ObjectBase* RankedPattern < SymbolType, RankType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace core {

template < class SymbolType, class RankType >
class SetConstraint< ::tree::RankedPattern < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::RankedPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		const ext::tree < common::ranked_symbol < SymbolType, RankType > > & m_content = pattern.getContent ( );
		return std::find(m_content.prefix_begin(), m_content.prefix_end(), symbol) != m_content.prefix_end() || pattern.template accessComponent < ::tree::SubtreeWildcard > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::RankedPattern < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::RankedPattern < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
	}
};

template < class SymbolType, class RankType >
class ElementConstraint< ::tree::RankedPattern < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::RankedPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return pattern.template accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::RankedPattern < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & symbol) {
		if( ( unsigned ) symbol.getRank() != 0 )
			throw ::tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );
	}
};

template < class SymbolType, class RankType >
struct normalize < tree::RankedPattern < SymbolType, RankType >, typename std::enable_if < ! std::is_same < tree::RankedPattern < SymbolType, RankType >, tree::RankedPattern < > >::value >::type > {
	static tree::RankedPattern < > eval ( tree::RankedPattern < SymbolType, RankType > && value ) {
		common::ranked_symbol < DefaultSymbolType, DefaultRankType > wildcard = alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( value ).getSubtreeWildcard ( ) );
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::tree < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > content = tree::TreeNormalize::normalizeRankedTree ( std::move ( value ).getContent ( ) );

		return tree::RankedPattern < > ( std::move ( wildcard ), std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace core */

#endif /* RANKED_PATTERN_H_ */
