/*
 * AdjacencyListUndirectedGraph.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "AdjacencyListUndirectedGraph.h"

#include <alib/compare>

#include <object/Object.h>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
#include "utils.h"

#include <registration/ValuePrinterRegistration.hpp>
#include <registration/XmlRegistration.hpp>

namespace graph {

AdjacencyListUndirectedGraph::AdjacencyListUndirectedGraph() : UndirectedGraph() {
}

AdjacencyListUndirectedGraph::AdjacencyListUndirectedGraph( const UndirectedGraph & graph ) {
	UndirectedGraph::fromUndirected( graph );
}

AdjacencyListUndirectedGraph::AdjacencyListUndirectedGraph( const DirectedGraph & dgraph ) {
	UndirectedGraph::fromDirected( dgraph );
}

GraphBase * AdjacencyListUndirectedGraph::clone() const {
	return new AdjacencyListUndirectedGraph( *this );
}

GraphBase * AdjacencyListUndirectedGraph::plunder() && {
	return new AdjacencyListUndirectedGraph( std::move( *this ) );
}

REPRESENTATION AdjacencyListUndirectedGraph::getRepresentation() const
{
	return REPRESENTATION::ADJACENCY_LIST;
}

std::string AdjacencyListUndirectedGraph::className() const
{
	return "AdjacencyListUndirectedGraph";
}

size_t AdjacencyListUndirectedGraph::nodeCount() const
{
	return adj.size();
}

size_t AdjacencyListUndirectedGraph::edgeCount() const
{
	return edges.size();
}

ext::set<Node> AdjacencyListUndirectedGraph::getNodes() const
{
	ext::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

ext::set<UndirectedEdge> AdjacencyListUndirectedGraph::getEdges() const
{
	return edges;
}

bool AdjacencyListUndirectedGraph::hasNode( const Node & node ) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyListUndirectedGraph::hasEdge( const UndirectedEdge & edge ) const
{
	return edges.find(edge) != edges.end();
}

bool AdjacencyListUndirectedGraph::hasEdge( const Node & first, const Node & second ) const
{
	auto search = adj.find(first);
	if (search == adj.end()) {
		return false;
	}
	return std::find(search->second.begin(), search->second.end(), second) != search->second.end();
}

ext::set<UndirectedEdge> AdjacencyListUndirectedGraph::findEdges( const Node & first, const Node & second) const
{
	ext::set<UndirectedEdge> out;

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, first, second)) {
			out.insert(e);
		}
	}

	return out;
}

ext::set<Node> AdjacencyListUndirectedGraph::neighbors( const Node & node ) const
{
	ext::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	std::copy(search->second.begin(), search->second.end(), std::inserter(out, out.begin()));
	return out;
}

ext::set<UndirectedEdge> AdjacencyListUndirectedGraph::neighborEdges( const Node & node ) const
{
	ext::set<UndirectedEdge> out;
	const ext::set<Node> &nghbrs = neighbors(node);

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, node, nghbrs)) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyListUndirectedGraph::addNode( const Node & node )
{
	if (adj.find(node) != adj.end())
		return false;

	adj[node]; // create empty list for this key
	return true;
}

bool AdjacencyListUndirectedGraph::addNode( const Node & node, int value )
{
	if (!addNode( node ))
		return false;

	UndirectedGraph::addNode( node, value );
	return true;
}

bool AdjacencyListUndirectedGraph::removeNode( const Node & node )
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);
	UndirectedGraph::removeNode(node);

	for ( auto iter = edges.begin ( ); iter != edges.end ( ); ) {
		if (iter->getFirstNode() == node) {
			listRemoveOne(adj[iter->getSecondNode()], node);
			iter = edges.erase(iter);
		} else if (iter->getSecondNode() == node) {
			listRemoveOne(adj[iter->getFirstNode()], node);
			iter = edges.erase(iter);
		} else {
			++ iter;
		}
	}

	return true;
}

bool AdjacencyListUndirectedGraph::addEdge( const UndirectedEdge & edge )
{
	if (edges.find(edge) != edges.end())
		return false;
	if (adj.find(edge.getFirstNode()) == adj.end())
		return false;
	if (adj.find(edge.getSecondNode()) == adj.end())
		return false;

	edges.insert(edge);

	adj[edge.getFirstNode()].push_back(edge.getSecondNode());
	adj[edge.getSecondNode()].push_back(edge.getFirstNode());

	return true;
}

bool AdjacencyListUndirectedGraph::addEdge( const UndirectedEdge & edge, int value )
{
	if (!addEdge( edge ))
		return false;

	UndirectedGraph::addEdge( edge, value );
	return true;
}

bool AdjacencyListUndirectedGraph::removeEdge( const UndirectedEdge & edge )
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);
	UndirectedGraph::removeEdge(edge);

	listRemoveOne(adj[edge.getFirstNode()], edge.getSecondNode());
	listRemoveOne(adj[edge.getSecondNode()], edge.getFirstNode());

	return true;
}

AdjacencyListUndirectedGraph AdjacencyListUndirectedGraph::parse( ext::deque<sax::Token>::iterator & input )
{
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, getXmlTagName());

	AdjacencyListUndirectedGraph graph;
	GraphFromXMLParser::parseNodes(input, graph);
	GraphFromXMLParser::parseUndirectedEdges(input, graph);
	GraphFromXMLParser::parseNodeValues(input, graph);
	GraphFromXMLParser::parseUndirectedEdgeValues(input, graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, getXmlTagName());
	return graph;
}

void AdjacencyListUndirectedGraph::compose( ext::deque<sax::Token> & out, const AdjacencyListUndirectedGraph & graph )
{
	out.emplace_back(getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeNodes(out, graph.getNodes());
	GraphToXMLComposer::composeEdges(out, graph.getEdges());
	GraphToXMLComposer::composeNodeValues(out, graph);
	GraphToXMLComposer::composeEdgeValues(out, graph);

	out.emplace_back(getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

object::ObjectBase* AdjacencyListUndirectedGraph::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} // namespace graph

namespace {

static auto valuePrinter = registration::ValuePrinterRegister < graph::AdjacencyMatrixUndirectedGraph > ( );

static auto xmlWrite = registration::XmlWriterRegister < graph::AdjacencyMatrixUndirectedGraph > ( );
static auto xmlRead = registration::XmlReaderRegister < graph::AdjacencyMatrixUndirectedGraph > ( );

static auto xmlGroup1 = registration::XmlRegisterTypeInGroup < graph::Graph, graph::AdjacencyMatrixUndirectedGraph > ( );
static auto xmlGroup2 = registration::XmlRegisterTypeInGroup < object::Object, graph::AdjacencyMatrixUndirectedGraph > ( );

} /* namespace */
