/*
 * UndirectedGraph.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "UndirectedGraph.h"
#include "AdjacencyListUndirectedGraph.h"
#include "AdjacencyMatrixUndirectedGraph.h"

#include <alib/iostream>
#include <alib/algorithm>
#include <sstream>
#include <stdexcept>

#include <sax/FromXMLParserHelper.h>
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
#include "../directed/DirectedGraph.h"
#include <object/Object.h>
#include <core/xmlApi.hpp>

namespace graph {


UndirectedGraph::UndirectedGraph() {
}

GRAPH_TYPE UndirectedGraph::getType() const
{
	return GRAPH_TYPE::UNDIRECTED;
}

void UndirectedGraph::fromUndirected( const UndirectedGraph & orig )
{
	int value;
	// convert nodes and their values
	for (const Node & n : orig.getNodes()) {
		try {
			value = orig.getNodeValue( n );
			this->addNode( n, value );
		} catch (const std::out_of_range & ex) { // original node doesn't have value
			this->addNode( n );
		}
	}
	// convert edges and their values
	for (const UndirectedEdge & e : orig.getEdges()) {
		try {
			value = orig.getEdgeValue( e );
			this->addEdge( UndirectedEdge( e.getFirstNode(), e.getSecondNode() ), value );
		} catch (const std::out_of_range & ex) { // original edge doesn't have value
			this->addEdge( UndirectedEdge( e.getFirstNode(), e.getSecondNode() ) );
		}
	}
	// convert embedding
	this->setEmbedding( orig.getEmbedding() );
}

void UndirectedGraph::fromDirected( const DirectedGraph & orig )
{
	int value;
	// convert nodes and their values
	for (const Node & n : orig.getNodes()) {
		try {
			value = orig.getNodeValue( n );
			this->addNode( n, value );
		} catch (const std::out_of_range & ex) { // original node doesn't have value
			this->addNode( n );
		}
	}
	// convert edges and their values
	for (const DirectedEdge & e : orig.getEdges()) {
		try {
			value = orig.getEdgeValue( e );
			this->addEdge( UndirectedEdge( e.getFromNode(), e.getToNode() ), value );
		} catch (const std::out_of_range & ex) { // original edge doesn't have value
			this->addEdge( UndirectedEdge( e.getFromNode(), e.getToNode() ) );
		}
	}
	// convert embedding
	this->setEmbedding( orig.getEmbedding() );
}

bool UndirectedGraph::addNode( const Node & node, int value )
{
	setNodeValue( node, value );
	return true;
}

bool UndirectedGraph::removeNode( const Node & node )
{
	return nodeValues.erase( node ) == 1;
}

bool UndirectedGraph::addEdge( const UndirectedEdge & edge, int value )
{
	setEdgeValue( edge, value );
	return true;
}

bool UndirectedGraph::removeEdge( const UndirectedEdge & edge )
{
	return edgeValues.erase( edge ) == 1;
}

bool UndirectedGraph::areNodesWeighted() const
{
	for (const Node & n : this->getNodes()) {
		if (nodeValues.find(n) == nodeValues.end())
			return false;
	}
	return true;
}

bool UndirectedGraph::areEdgesWeighted() const
{
	for (const UndirectedEdge & e : this->getEdges()) {
		if (edgeValues.find(e) == edgeValues.end())
			return false;
	}
	return true;
}

int UndirectedGraph::getNodeValue( const Node & node ) const
{
	return nodeValues.at( node );
}

void UndirectedGraph::setNodeValue( const Node & node, int value )
{
	nodeValues.insert( {node, value} );
}

int UndirectedGraph::getEdgeValue( const UndirectedEdge & edge ) const
{
	return edgeValues.at( edge );
}

void UndirectedGraph::setEdgeValue( const UndirectedEdge & edge, int value )
{
	edgeValues.insert( {edge, value} );
}

ext::unordered_map< Node, ext::vector< Node > > UndirectedGraph::getEmbedding() const
{
	return m_embedding;
}

void UndirectedGraph::setEmbedding( const ext::unordered_map< Node, ext::vector< Node > > embedding )
{
	this->m_embedding = embedding;
}

int UndirectedGraph::compare( const UndirectedGraph & other ) const
{
	ext::set<Node> thisNodes = this->getNodes();
	ext::set<UndirectedEdge> thisEdges = this->getEdges();
	ext::set<Node> otherNodes = other.getNodes();
	ext::set<UndirectedEdge> otherEdges = other.getEdges();

	auto first = ext::tie(thisNodes, thisEdges, nodeValues, edgeValues, m_embedding);
	auto second = ext::tie(otherNodes, otherEdges, other.nodeValues, other.edgeValues, other.m_embedding);

	static ext::compare<decltype(first)> comp;
	return comp(first, second);
}

void UndirectedGraph::operator>>( std::ostream & out ) const
{
	out << "(UndirectedGraph " << this->toString() << ")";
}

UndirectedGraph::operator std::string() const
{
	return this->toString();
}

std::string UndirectedGraph::toString() const
{
	std::stringstream ss;
	for (const Node & node : getNodes()) {
		ss << node;
	}
	for (const UndirectedEdge & edge : getEdges()) {
		ss << edge;
	}
	return ss.str();
}
/*
UndirectedGraph * UndirectedGraph::parse(ext::deque<sax::Token>::iterator & input)
{
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, UndirectedGraph::getXmlTagName());

	UndirectedGraph * graph;
	switch (GraphFromXMLParser::parseRepresentation(input)) {
	 case REPRESENTATION::ADJACENCY_LIST:
		graph = new AdjacencyListUndirectedGraph();
		break;
	 case REPRESENTATION::ADJACENCY_MATRIX:
		graph = new AdjacencyMatrixUndirectedGraph();
		break;
	 case REPRESENTATION::INCIDENCE_MATRIX:

	 default:
		graph = NULL;
		break;
	}
	GraphFromXMLParser::parseNodes(input, *graph);
	GraphFromXMLParser::parseUndirectedEdges(input, *graph);
	GraphFromXMLParser::parseNodeValues(input, *graph);
	GraphFromXMLParser::parseUndirectedEdgeValues(input, *graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, UndirectedGraph::getXmlTagName());
	return graph;
}

void UndirectedGraph::compose(ext::deque<sax::Token> & out) const
{
	out.emplace_back(UndirectedGraph::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeRepresentation(out, representationToString(this->getRepresentation()));
	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(UndirectedGraph::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}
*/

} // namespace graph

/*
namespace core {

auto undirectedGraphParserRegister = xmlApi<graph::Graph>::ParserRegister<graph::UndirectedGraph>();
auto undirectedGraphParserRegister2 = xmlApi<object::Object>::ParserRegister<graph::UndirectedGraph>();

} // namespace core
*/
