/*
 * XmlDataFactory.hpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Jan Travnicek
 */

#ifndef XML_DATA_FACTORY_HPP_
#define XML_DATA_FACTORY_HPP_

#include <alib/string>
#include <alib/deque>
#include <measure>

#include <sax/Token.h>
#include <sax/SaxParseInterface.h>
#include <sax/SaxComposeInterface.h>
#include <core/xmlApi.hpp>
#include <exception/xml/CommonException.h>
#include <measure/xml/MeasurementResults.hpp>

namespace factory {

/**
 * Data builder.
 */
class XmlDataFactory {
public:

	class fromFile {
		const std::string & filename;

	public:
		fromFile ( const std::string & file ) : filename ( file ) {
		}

		/**
		 * Parses the XML in file and returns the String.
		 * @param filename path to the file
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			ext::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseFile ( filename, tokens );

			return fromTokens ( std::move( tokens ) );
		}
	};

	class fromString {
		const std::string & string;

	public:
		fromString ( const std::string & str ) : string ( str ) {
		}

		/**
		 * Parses the XML and returns the String.
		 * @param str string containing the XML
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			ext::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseMemory ( string, tokens );

			return fromTokens ( std::move ( tokens ) );
		}
	};

	class fromStdin {
	public:
		/**
		 * Parses the XML from stdin and returns the String.
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			ext::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseStdin ( tokens );

			return fromTokens ( std::move ( tokens ) );
		}
	};

	class fromStream {
		std::istream & in;

	public:
		fromStream ( std::istream & i ) : in ( i ) {
		}

		/**
		 * Parses the XML from stream and returns the String.
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			ext::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseStream ( in, tokens );

			return fromTokens ( std::move ( tokens ) );
		}
	};

	class fromTokens {
		ext::deque < sax::Token > tokens;

	public:
		fromTokens ( ext::deque < sax::Token > && toks ) : tokens ( std::move ( toks ) ) {
		}

		/**
		 * Parses the String from list of tokens.
		 * @param tokens XML represented as list of tokens
		 * @return parsed String
		 */
		template < class T >
		operator T ( ) {
			core::xmlApiInputContext context ( tokens.begin ( ) );

			if ( context == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

			if ( core::xmlApi < exception::CommonException >::first ( context ) ) throw core::xmlApi < exception::CommonException >::parse ( context );

			measurements::start ( "XML Parser", measurements::Type::INIT );
			T res = core::xmlApi < T >::parse ( context );
			measurements::end ( );

			if ( context != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

			return res;
		}
	};

	/**
	 * Determines whether the input token stream contains given type
	 * @param tokens
	 * @return bool
	 */
	template < class T >
	static bool first ( const ext::deque < sax::Token > & tokens ) {
		if( tokens.empty ( ) ) throw exception::CommonException ( "Empty tokens list" );

		return core::xmlApi < exception::CommonException >::first ( tokens.begin ( ) ) || core::xmlApi < T >::first ( tokens.begin ( ) );
	}

	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template < class T >
	static void toFile ( const T & data, const std::string & filename ) {
		ext::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printFile ( filename, tokens );
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template < class T >
	static std::string toString ( const T & data ) {
		ext::deque < sax::Token > tokens = toTokens < T > ( data );
		std::string str;
		sax::SaxComposeInterface::printMemory ( str, tokens );

		return str;
	}

	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template < class T >
	static void toStdout ( const T & data ) {
		ext::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printStdout ( tokens );
	}

	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template < class T >
	static void toStream ( const T & data, std::ostream & out ) {
		ext::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printStream ( out, tokens );
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template < class T >
	static ext::deque < sax::Token > toTokens ( const T & data ) {
		core::xmlApiOutputContext context;

		measurements::start ( "XML Composer", measurements::Type::FINALIZE );
		core::xmlApi < T >::compose ( context, data );
		measurements::end ( );

		return context;
	}

};

template < >
inline bool XmlDataFactory::first < exception::CommonException > ( const ext::deque < sax::Token > & tokens ) {
	return core::xmlApi < exception::CommonException >::first ( tokens.begin ( ) );
}

template < >
inline XmlDataFactory::fromTokens::operator exception::CommonException ( ) {
	core::xmlApiInputContext context ( tokens.begin ( ) );

	if ( context == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

	exception::CommonException res = core::xmlApi < exception::CommonException >::parse ( context );

	if ( context != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return res;
}

template < >
inline XmlDataFactory::fromTokens::operator measurements::MeasurementResults ( ) {
	core::xmlApiInputContext context ( tokens.begin ( ) );

	if ( context == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

	if ( core::xmlApi < exception::CommonException >::first ( context ) ) throw core::xmlApi < exception::CommonException >::parse ( context );

	measurements::MeasurementResults res = core::xmlApi < measurements::MeasurementResults >::parse ( context );

	if ( context != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return res;
}

template < >
inline ext::deque < sax::Token > XmlDataFactory::toTokens ( const measurements::MeasurementResults & data ) {
	core::xmlApiOutputContext context;

	core::xmlApi < measurements::MeasurementResults >::compose ( context, data );

	return context;
}

} /* namespace factory */

#endif /* XML_DATA_FACTORY_HPP_ */
